from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

class Processor:
    def process(self, item):
        raise NotImplementedError('Subclass must implement method')
